<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
  <head>
    <meta http-equiv="Content-Type" content="text/html; charset=utf-8"/>
  </head>
  <body>
    <h1>Logout error</h1>
    <p>Your Identity Provider ask this system to do a global logout but your federated session is lost.</p>

    <p>Even if your local session in this system has been closed, you have probably open sessions in other systems.</p>

    <p>In order to prevent illicit use of your personal information, please close your browser window and/or remove your cookies from your browser.</p>

    <p>Sorry for this inconvenience.</p>
  </body>
</html>
